/* vi: set sw=4 ts=4: */
/*
 * stats collection
 *
 * Copyright (C) 2018 Grandstream Networks, Inc.
 *
 * Licensed under GPLv2, see file LICENSE in this source tree.
 */

//kbuild:lib-$(CONFIG_STATS)     += stats.o

//config:config STATS
//config:	bool "stats"
//config:	default n
//config:	help
//config:	  grandstream stats collection

//usage:#define stats_trivial_usage
//usage:       "..."
//usage:#define stats_full_usage "\n\n"
//usage:       "collect stats for grandstream device"
//usage:
//usage:#define cat_example_usage
//usage:       "$ stats\n"

#include "libbb.h"
#include "nvram.h"
#include <syslog.h>

static void read_file(const char *file, char *p, int len) {
    int fd;

    fd = xopen( file, O_RDONLY );
    read( fd, p, len );
    xclose( fd );
}

static void concat_syslog(void) {
    DIR *dir;
    struct dirent *entry;
    int fout;

    xchdir( "/var/user" );
    dir = xopendir( "./" );
    
    fout = xopen( "syslog_output.txt", O_RDWR | O_TRUNC | O_CREAT );

    while ( ( entry = readdir( dir ) ) ) {
        if ( strncmp( entry->d_name, "syslog_", 7 ) == 0 ) {
            int fin = xopen( entry->d_name, O_RDONLY );
            bb_copyfd_eof( fin, fout );
            xclose( fin );
        }
    }

    closedir( dir );
    xclose( fout );
}

static void get_statinfo(const char *image, unsigned int *rss, unsigned long *utime, unsigned long *stime) {
    DIR *dir;
    struct dirent *entry;
    struct stat statbuf;
    int fin;
    char buffer[2048];

    xchdir( "/proc/" );
    dir = xopendir( "./" );

#define RSS_INDEX 24
#define UTIME_INDEX 14
#define STIME_INDEX 15
#define IMAGE_INDEX 2

    while ( ( entry = readdir( dir ) ) ) {
        if ( lstat( entry->d_name, &statbuf ) != 0 )
            continue;

        if ( S_ISDIR( statbuf.st_mode ) ) {
            char *p;
            int index = 1;

            for ( p = entry->d_name; *p; p++ ) {
                if ( !isdigit( *p ) )
                    break;
            }
            if ( *p )
                continue;

            fin = open( "stat", O_RDONLY );
            if ( fin < 0 )
                continue;

            read( fin, buffer, sizeof ( buffer ) );
            xclose( fin );

            p = buffer;
            while ( index != IMAGE_INDEX ) {
                for ( p = buffer; *p != ' '; p++ );
                index++;
            }
            if ( strncmp( image, p, strlen( image ) ) != 0 ) {
                continue;
            }

            while ( index != UTIME_INDEX ) {
                for ( p = buffer; *p != ' '; p++ );
                index++;
            }
            *utime = bb_strtoul( ++p, NULL, 10 );

            while ( index != STIME_INDEX ) {
                for ( p = buffer; *p != ' '; p++ );
                index++;
            }
            *stime = bb_strtoul( ++p, NULL, 10 );

            while ( index != RSS_INDEX ) {
                for ( p = buffer; *p != ' '; p++ );
                index++;
            }
            *rss = (int)bb_strtoul( ++p, NULL, 10 );
            closedir( dir );
            return;
        }
    }

    closedir( dir );

    *rss = 0;
    *utime = 0;
    *stime = 0;
}

int stats_main(int argc, char **argv)
{
    int fin;
    char *p;
    int count = 0;
    char buffer[32];
    char log[2048];
    char prefix[32];

    struct {
        char model[16];
        char mac[16];
        char version[16];
        char wan[16];
    } device;

    struct {
        char log_level[16];
        char enabled_syslog[16];
    } config;


    struct {
        char load[32];
        struct {
            unsigned long rx;
            unsigned long tx;
            unsigned long rx_delta;
            unsigned long tx_delta;
            unsigned long rx_last;
            unsigned long tx_last;
        } net;
        struct {
            unsigned long utime;
            unsigned long stime;
            unsigned long utime_last;
            unsigned long stime_last;
            unsigned int cpu_percent;
            unsigned int rss;
        } gs_gui;
        struct {
            unsigned long utime;
            unsigned long stime;
            unsigned long utime_last;
            unsigned long stime_last;
            unsigned int cpu_percent;
            unsigned int rss;
        } gs_phone;
        struct {
            unsigned long utime;
            unsigned long stime;
            unsigned long utime_last;
            unsigned long stime_last;
            unsigned int cpu_percent;
            unsigned int rss;
        } app_dsp;
        struct {
            unsigned long utime;
            unsigned long stime;
            unsigned long utime_last;
            unsigned long stime_last;
            unsigned int cpu_percent;
            unsigned int rss;
        } gs_cpe;
        unsigned long cpu;
        unsigned long cpu_last;
    } data = { 0 };

    int user_space_write = 0;
    int backoff = 0;
    
    printf("%s:%d\n", __func__, __LINE__);
    
    read_file( "/proc/gxp/dev_info/dev_alias", device.model, sizeof ( device.model ) );
    read_file( "/proc/gxp/dev_info/dev_mac", device.mac, sizeof ( device.mac ) );
    nvram_get_safe( "68", device.version, sizeof ( device.version ) );
    nvram_get_safe( "wan_device", device.wan, sizeof ( device.wan ) );
    nvram_get_safe( "208", config.log_level, sizeof ( config.log_level ) );
    nvram_get_safe( "debug_log", config.enabled_syslog, sizeof ( config.enabled_syslog ) );

    sprintf( prefix, "%s_STAT", device.model );
    openlog( prefix, 0, LOG_DAEMON );

    printf("%s:%d\n", __func__, __LINE__);

    concat_syslog();
    
    printf("%s:%d\n", __func__, __LINE__);

    while ( 1 ) {
        sleep(5);

        /* load average */
        fin = xopen( "/proc/loadavg", O_RDONLY );
        read( fin, data.load, sizeof ( data.load ) );
        xclose( fin );

        for ( p = data.load; *p && *p != ' '; p++ );
        for ( ; *p && *p != ' '; p++ );
        for ( ; *p && *p != ' '; p++ );
        *p = 0;

        count++;
        if ( atoi( data.load ) > 0 ) {
            if ( count == 5 ) {
                count = 0;
                backoff = 0;
            }
            else if ( backoff ) {
                continue;
            }
            else {
                count = 0;
                backoff = 1;
            }
        }
        else if ( backoff ) {
            backoff = 0;
        }
        else if ( count == 5 ) {
            count = 0;
        }
        else {
            continue;
        }

        /* net stats */
#define NET_PATH "/sys/devices/platform/6e00000.ethernet/net/eth0/statistics/"
        fin = xopen( NET_PATH "rx_bytes", O_RDONLY );
        read( fin, buffer, sizeof ( buffer ) );
        xclose( fin );
        data.net.rx = atoi( buffer );

        fin = xopen( NET_PATH "tx_bytes", O_RDONLY );
        read( fin, buffer, sizeof ( buffer ) );
        xclose( fin );
        data.net.tx = atoi( buffer );

        data.net.rx_delta = data.net.rx - data.net.rx_last;
        data.net.tx_delta = data.net.tx - data.net.tx_last;
        data.net.tx_last = data.net.tx;
        data.net.rx_last = data.net.rx;

        /* cpu stats */
        fin = xopen( "/proc/stat", O_RDONLY );
        read( fin, buffer, sizeof ( buffer ) );
        xclose( fin );

        data.cpu = 0;
        data.cpu += bb_strtoul( buffer, &p, 10 );
        data.cpu += bb_strtoul( p, &p, 10 );
        data.cpu += bb_strtoul( p, &p, 10 );
        data.cpu += bb_strtoul( p, &p, 10 );

        get_statinfo( "(gs_gui)", &data.gs_gui.rss, &data.gs_gui.utime, &data.gs_gui.stime );
        get_statinfo( "(gs_phone)", &data.gs_phone.rss, &data.gs_phone.utime, &data.gs_phone.stime );
        get_statinfo( "(app_dsp)", &data.app_dsp.rss, &data.app_dsp.utime, &data.app_dsp.stime );

        data.gs_gui.rss *= 4096;
        data.gs_phone.rss *= 4096;
        data.app_dsp.rss  *= 4096;

#define CPU_PERCENT(x) (int) ( ( ( ( (x).utime + (x).stime ) - ( (x).utime_last + (x).stime_last ) ) /  \
            (float)( data.cpu - data.cpu_last ) ) * 100 );
        data.gs_gui.cpu_percent = CPU_PERCENT( data.gs_gui );
        data.gs_phone.cpu_percent = CPU_PERCENT( data.gs_phone );
        data.app_dsp.cpu_percent = CPU_PERCENT( data.app_dsp );

        syslog( LOG_INFO, "[%s][%s] [ %s | %lu %lu | %d | %d | %d | %d | %d | %d ]\n",
                 device.mac, device.version,
                 data.load,
                 data.net.tx_delta, data.net.rx_delta,
                 data.gs_gui.rss, data.gs_gui.cpu_percent,
                 data.gs_phone.rss, data.gs_phone.cpu_percent,
                 data.app_dsp.rss, data.app_dsp.cpu_percent );
        printf(  "[%s][%s] [ %s | %lu %lu | %d | %d | %d | %d | %d | %d ]\n",
                 device.mac, device.version,
                 data.load,
                 data.net.tx_delta, data.net.rx_delta,
                 data.gs_gui.rss, data.gs_gui.cpu_percent,
                 data.gs_phone.rss, data.gs_phone.cpu_percent,
                 data.app_dsp.rss, data.app_dsp.cpu_percent );

        if ( config.enabled_syslog ) {
            struct tm tm_time;
            struct timespec ts;
            FILE *pf;
            struct {
                char total[64];
                char free[64];
                char available[64];
                char buffers[64];
                char cached[64];
            } mem;
            /* deleting the oldest log every day should be a cron job */
        
            get_statinfo( "(gs_cpe)", &data.gs_cpe.rss, &data.gs_cpe.utime, &data.gs_cpe.stime );
            data.gs_cpe.rss *= 4096;
            data.gs_cpe.cpu_percent = CPU_PERCENT( data.gs_cpe );

            pf = fopen( "/proc/meminfo", "r" );
            fgets( mem.total, sizeof ( mem.total ), pf );
            fgets( mem.free, sizeof ( mem.free ), pf );
            fgets( mem.available, sizeof ( mem.available ), pf );
            fgets( mem.buffers, sizeof ( mem.buffers ), pf );
            fgets( mem.cached, sizeof ( mem.cached ), pf );
            fclose( pf );

            /* lifted from date applet */
#if ENABLE_FEATURE_DATE_NANO
            syscall(__NR_clock_gettime, CLOCK_REALTIME, &ts);
#else
            time(&ts.tv_sec);
#endif
            strftime( buffer, sizeof ( buffer ), "/var/user/syslog_%m_%d_%Y", localtime_r( &ts.tv_sec, &tm_time ) );

            fin = xopen( buffer, O_RDWR | O_APPEND );

            strftime( buffer, sizeof ( buffer ), "%m/%d/%Y %T", localtime_r( &ts.tv_sec, &tm_time ) );
            sprintf( log, "%s> g: [%d]%d | p: [%d]%d | d: [%d]%d | c: [%d]%d => Mem: %s %s %s %s %s\n",
                     buffer,
                     data.gs_gui.rss, data.gs_gui.cpu_percent,
                     data.gs_phone.rss, data.gs_phone.cpu_percent,
                     data.app_dsp.rss, data.app_dsp.cpu_percent,
                     data.gs_cpe.rss, data.gs_cpe.cpu_percent,
                     mem.total, mem.free, mem.available, mem.buffers, mem.cached );
            write( fin, buffer, strlen( buffer ) );
            xclose( fin );

            user_space_write++;
            if ( user_space_write > 10 ) {
                nvram_set( ":user_space_dirty", "1" );
                user_space_write = 0;
            }
        }
       
        data.net.rx_last = data.net.rx;
        data.net.tx_last = data.net.tx;
        data.gs_gui.utime_last = data.gs_gui.utime;
        data.gs_gui.stime_last = data.gs_gui.stime;
        data.gs_phone.utime_last = data.gs_phone.utime;
        data.gs_phone.stime_last = data.gs_phone.stime;
        data.app_dsp.utime_last = data.app_dsp.utime;
        data.app_dsp.stime_last = data.app_dsp.stime;
        data.gs_cpe.utime_last = data.gs_cpe.utime;
        data.gs_cpe.stime_last = data.gs_cpe.stime;
        data.cpu_last = data.cpu;
    }
    return 0;
}
