/*
 * NVRAM variable manipulation
 *
 * Copyright 2005, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id: nvram.h,v 1.1 2013/05/10 18:59:49 jvan Exp $
 *
 * SCF - ADM5120 port - 6/30/2006
 *     - removed broadcom specific portions and OS indepent layer
 *     - removed MTD dependence, ported to raw flash access routines
 *     - added zlib support and checksumming for nvram file
 *                     
 */

#ifndef _bcmnvram_h_
#define _bcmnvram_h_

#define NVRAM_DEV_MAJOR 241
#define NVRAM_DEV_MINOR 0

#define MARK_TEMP         1
#define MARK_L1LOCKED     2
#define MARK_L2LOCKED     3

#define ROYALTY_START    9600
#define ROYALTY_END      9899

// Process owner
typedef enum
{
    PROCESS_SYSTEM = 0, 
    PROCESS_WEB, 
    PROCESS_TELNET,
    PROCESS_GUI,
    PROCESS_PHONE,
    PROCESS_SERVICE,
    PROCESS_CPE,
    PROCESS_UNREGISTERED
} PROCESS_OWNER;

#ifndef _LANGUAGE_ASSEMBLY

#include <linux/types.h>
#include <stdbool.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

struct nvram_header {
	int magic;
	int len;
	int space_used;
	char reserved[10];
};

struct nvram_tuple {
	char *name;
	char *value;
	struct nvram_tuple *next;
};

/*
 * we don't know how much data we'll get out of the compressed data stored in flash,
 * so we'll create a linked list element for each call to zlib_inflate with 
 * the number of bytes available in the decompression buffer equal to the size of the
 * array pointed to by *data, then glue all the array elements together for processing
 */
struct nvram_inflate_stream {
	char *data;
	struct nvram_inflate_stream *next;
};

/*
 * Initialize NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern int nvram_init(void*);

/*
 * Disable NVRAM access. May be unnecessary or undefined on certain
 * platforms.
 */
extern void nvram_exit(void);

/*
 * Get the value of an NVRAM variable. The pointer returned may be
 * invalid after a set.
 * @param	name	name of variable to get
 * @return	value of variable or NULL if undefined
 */
extern char * nvram_get(const char *name);

/*
 * Race free version of nvram_get
 * @param name		name of variable to get
 * @param buffer	buffer to store result
 * @param length	length of buffer
 */
extern char * nvram_get_safe( const char * name, char * buffer, const unsigned int length );

/*
 * Race free version of nvram_get
 * @param name		name of variable to get
 * @param buffer	buffer to store result
 * @param length	length of buffer
 */
extern char * nvram_get_safe_security( const char * name, char * buffer, const unsigned int length );

/* 
 * Get the value of an NVRAM variable.
 * @param	name	name of variable to get
 * @return	value of variable or NUL if undefined
 */
#define nvram_safe_get(name) (nvram_get(name) ? : "")

/*
 * Match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is string equal
 *		to match or FALSE otherwise
 */
static __inline__ int
nvram_match(char *name, char *match) {
	const char *value = nvram_get(name);
	return (value && !strcmp(value, match));
}

/*
 * Inversely match an NVRAM variable.
 * @param	name	name of variable to match
 * @param	match	value to compare against value of variable
 * @return	TRUE if variable is defined and its value is not string
 *		equal to invmatch or FALSE otherwise
 */
static __inline__ int
nvram_invmatch(char *name, char *invmatch) {
	const char *value = nvram_get(name);
	return (value && strcmp(value, invmatch));
}

/*
 * Set the value of an NVRAM variable. The name and value strings are
 * copied into private storage. Pointers to previously set values
 * may become invalid. The new value may be immediately
 * retrieved but will not be permanently stored until a commit.
 * @param	name	name of variable to set
 * @param	value	value of variable
 * @return	0 on success and errno on failure
 */
extern int nvram_set( const char *name, const char *value );
extern int nvram_set_security( const char *name, const char *value );

/*
 * Unset an NVRAM variable. Pointers to previously set values
 * remain valid until a set.
 * @param	name	name of variable to unset
 * @return	0 on success and errno on failure
 * NOTE: use nvram_commit to commit this change to flash.
 */
extern int nvram_unset( const char *name );
extern int nvram_unset_security( const char *name );

/*
 * Commit NVRAM variables to permanent storage. All pointers to values
 * may be invalid after a commit.
 * NVRAM values are undefined after a commit.
 * @return	0 on success and errno on failure
 */
extern int nvram_commit(void);
extern int nvram_commit_sync(void);
extern int nvram_commit_override(void); /* nvram_commit_sync overriding rate-limiting mechanism */

/*
 * Get all NVRAM variables (format name=value\0 ... \0\0).
 * @param	buf	buffer to store variables
 * @param	count	size of buffer in bytes
 * @return	0 on success and errno on failure
 */
extern int nvram_getall(char *buf, int count);

extern int nvram_space_used( void );
extern int nvram_ram_used( void );
extern int nvram_erase_all( void );
extern int nvram_erase_all_security( void );

extern int nvram_set_notification( const char *name ) ;
extern int nvram_unset_notification( const char *name ) ;
extern int nvram_check_notification( const char *name ) ;

/*
 * OPTIONAL: by setting originate, it will help us determine 
 * the process which request/submit pvalue.
 * default: 0 (ORIGINATED_FROM_SYS)[5/3/2013 Art]
 */
extern void nvram_set_originator( PROCESS_OWNER originator );

// allow other process to pull the recent change list [5/1/2013 Art]
extern char * nvram_pull_logs(char * buffer, unsigned int length);
extern char * nvram_peek_logs(unsigned short startId, char* buf, unsigned int length);
// return whether log is full
extern int nvram_check_log_full(void);
// return whether log is empty 
extern int nvram_check_log_empty(void);
// return current log id
extern int nvram_check_log_current(void);

extern int nvram_subscribe_updates( const char* component, const char* values);
extern char* nvram_pull_subscribed_updates( const char* component, char* buffer, unsigned int length );

extern char* nvram_l2get_safe( const char *name, char *buffer, unsigned int length );
extern char* nvram_l2get_safe_security( const char *name, char *buffer, unsigned int length );
extern int nvram_l1set( const char *name, const char *value );
extern int nvram_l1set_security( const char *name, const char *value );
extern int nvram_l2set( const char *name, const char *value );
extern int nvram_l2set_security( const char *name, const char *value );
extern int nvram_l1lock( const char *name );
extern int nvram_l1lock_security( const char *name );
extern int nvram_l2lock( const char *name );
extern int nvram_l2lock_security( const char *name );

#endif /* _LANGUAGE_ASSEMBLY */

#ifdef __cplusplus
}
#endif

#define NVRAM_MAGIC		    0x48534C46	/* 'FLSH' */
#define NVRAM_SPACE_USED	0x00000001
#define NVRAM_RAM_USED		0x00000002
#define NVRAM_COMMIT_SYNC   0x00000004
#define NVRAM_ERASE_ALL	    0x00000010
#define NVRAM_BUFFER_SPACE  (0x2000*32)

#endif /* _bcmnvram_h_ */
